/*******************************************************************************
 * Copyright 2024 ModalAI Inc.
 ******************************************************************************/

/**
 * @file       mcv-test-ft.cpp
 *
 * @brief      Tests the feature tracking process
 *
 * @author     Kyle Tyni (kyle.tyni@modalai.com)
 * @date       2024
 */


#include <iostream>

#include <modal_pipe.h>
#include <modalcv.h>
#include <opencv2/imgcodecs.hpp>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>


// called whenever we connect or reconnect to the server
static void _connect_cb(__attribute__((unused)) int ch,
                        __attribute__((unused)) void *context) {
    printf("connected\n");
    return;
}

// called whenever we disconnect from the server
static void _disconnect_cb(__attribute__((unused)) int ch,
                           __attribute__((unused)) void *context) {
    printf("disconnected\n");
    return;
}

typedef struct mcv_cvp_vft_metadata {
    int16_t num_of_features;
    int32_t frame_id;
    int64_t frame_timestamp_ns;
} mcv_cvp_vft_metadata;


int16_t num_outputs;
mcv_cvp_feature_tracker_handle handle;
mcv_cvp_feature_tracker_output_t output[MAX_N_FEATURES];
mcv_cvp_vft_metadata frame_metadata;
int16_t frame_count = 0;

// holds pointers to metadata struct and array struct
const void* bufs[2];
size_t lens[2];

static void _camera_cb(__attribute__((unused)) int ch, 
                       camera_image_metadata_t meta, char *frame,
                       __attribute__((unused)) void *context) {

    // how to validate the data??
    if (meta.magic_number != CAMERA_MAGIC_NUMBER) {
        printf("ERROR receiving camera image data from pipe...\n");
        return;
    }


    // should frame data be stored in some buffer somewhere
    mcv_cvp_feature_tracker_process(0, handle, (uint8_t*)frame,
                                    &num_outputs, output);


    printf("frame_count: %4d  features: %4d   bytes: %5d\n", frame_count, 
           num_outputs, sizeof(mcv_cvp_feature_tracker_output_t) * num_outputs);           
    frame_count++;
    

    // publish the data
    frame_metadata.num_of_features = num_outputs;
    frame_metadata.frame_id = meta.frame_id;
    frame_metadata.frame_timestamp_ns = meta.timestamp_ns;

    bufs[0] = (const void*)&frame_metadata;
    bufs[1] = (const void*)output;

    lens[0] = sizeof(mcv_cvp_vft_metadata);
    lens[1] = sizeof(mcv_cvp_feature_tracker_output_t) * num_outputs;

    pipe_server_write_list(0, 2, bufs, lens);

    return;
}

const char* CLIENT_NAME  = "mcv-test-ft";
const char* IN_PIPE_NAME = "tracking_front";
const char* SERVER_NAME = "vft-output-server";

pipe_info_t out_pipe_info;

int main(int argc, char* argv[]) {


/*  TODO: image dimensions are hardcoded, is there a way to make them dynamic on
    runtime  */

    int w = 1280, h = 800;


    // mcv feature tracker config init

    // configuration object for feature tracking
    mcv_cvp_feature_tracker_config_t feature_tracker_config;

    // which octave of the pyramid to perform optic flow on
    feature_tracker_config.opt_flow_octave = 2;

    // create a binary mask for features (NULL implies all 1's)
    cv::Mat mask = cv::Mat::ones(h, w, CV_8UC1) * 255;
    cv::Rect lower_right_corner(w / 2, h / 2, w / 2, h / 2);
    mask(lower_right_corner).setTo(cv::Scalar(0));
    feature_tracker_config.feature_mask = mask.data;

    // set up feature point extraction
    feature_tracker_config.pyr_fpx_config.width = w;
    feature_tracker_config.pyr_fpx_config.height = h;
    feature_tracker_config.pyr_fpx_config.n_octaves = 3;
    feature_tracker_config.pyr_fpx_config.n_scales_per_octave = 1;
    feature_tracker_config.pyr_fpx_config.mode = MCV_CVP_FPX_ZONE;
    feature_tracker_config.pyr_fpx_config.nms_mode = MCV_CVP_FPX_5_TAP_NMS;
    feature_tracker_config.pyr_fpx_config.score_threshold = 500;
    feature_tracker_config.pyr_fpx_config.robustness = 10;
    feature_tracker_config.pyr_fpx_config.output_ubwc = 1;
    feature_tracker_config.pyr_fpx_config.zone_dims[0] = 16;
    feature_tracker_config.pyr_fpx_config.zone_dims[1] = 8;
    feature_tracker_config.pyr_fpx_config.zone_dims[2] = 4;
    // memset(feature_tracker_config.pyr_fpx_config.n_filter_coeffs, 3,
    //        MCV_CVP_MAX_PYRAMID_OCTAVES * sizeof(uint32_t));
    for (int i = 0; i < (int)feature_tracker_config.pyr_fpx_config.n_octaves; i++)
        feature_tracker_config.pyr_fpx_config.enable_fpx[i] = 1;

    // set up opt flow
    // width/height here are dictated by the octave we want to do flow on as defined
    // above in the feature_tracker_config
    feature_tracker_config.opt_flow_config.width =
        w / pow(2, feature_tracker_config.opt_flow_octave);
    feature_tracker_config.opt_flow_config.height =
        h / pow(2, feature_tracker_config.opt_flow_octave);
    feature_tracker_config.opt_flow_config.en_stats = 0;
    feature_tracker_config.opt_flow_config.input_ubwc =
        1;  // pyrfpx output ubwc

    // get a handle for our patch tracker
    handle = mcv_cvp_feature_tracker_init(feature_tracker_config);
    if (NULL == handle) {
        fprintf(stderr, "ERROR getting patch tracker handle\n");
        return -1;
    }


    enable_signal_handler();


    // Client pipe for getting camera image data
    pipe_client_set_camera_helper_cb(0, _camera_cb, NULL);
    pipe_client_set_connect_cb(0, _connect_cb, NULL);
    pipe_client_set_disconnect_cb(0, _disconnect_cb, NULL);

    pipe_info_t pipe_info;

    if (pipe_exists(IN_PIPE_NAME)) {
        
        pipe_get_info(IN_PIPE_NAME, &pipe_info);
        
        printf("name: %s\n", pipe_info.name);
        printf("server_name: %s\n", pipe_info.server_name);
        printf("type: %s\n", pipe_info.type);
        printf("location: %s\n", pipe_info.location);
    }

    int ret = 0;

    ret = pipe_client_open(0, IN_PIPE_NAME, CLIENT_NAME,
                           EN_PIPE_CLIENT_CAMERA_HELPER,
                           sizeof(camera_image_metadata_t));
    if (ret) {
        fprintf(stderr, "ERROR: failed to open pipe %s\n", IN_PIPE_NAME);
        pipe_print_error(ret);
        return ret;
    }

    // check for errors trying to connect to the server pipe
    if (ret < 0) {
        pipe_print_error(ret);
        return -1;
    }


    // Server pipe for publishing feature tracker data

    // pipe server info
    strcpy(out_pipe_info.name, "vft_output");
    strcpy(out_pipe_info.location, "/run/mpa/vft_output/");
    strcpy(out_pipe_info.type, "mcv_cvp_feature_tracker_output");
    strcpy(out_pipe_info.server_name, SERVER_NAME);
    out_pipe_info.size_bytes = MODAL_PIPE_DEFAULT_PIPE_SIZE;
    out_pipe_info.server_pid = 1337;


    int ch = pipe_server_get_next_available_channel();

    ret = pipe_server_create(ch, out_pipe_info, 0);

    if (ret) {
        fprintf(stderr, "ERROR: failed to create server %s\n", SERVER_NAME);
        pipe_print_error(ret);
        return ret;
    }

    make_pid_file(SERVER_NAME);


    // keep going until the signal handler sets the running flag to 0
    main_running = 1;
    
    while (main_running) {
        usleep(500000);
    } 


    // shut down clients and pipes
    pipe_client_close_all();
    pipe_server_close_all();
    remove_pid_file(SERVER_NAME);

    return 0;
}

